/*
** fixup.c for elfsh
** 
** Started on  Fri Jul 27 04:56:06 2001 mayhem
** Last update Sat May 17 09:25:27 2003 mayhem
*/
#include "libelfsh.h"



/* TERMINAL FUNCTION ! Fixup size for section symbols, if symtab doesnt exist, create it */
elfshsect_t	*elfsh_fixup_symtab(elfshobj_t *file, int *strindex)
{
  elfshsect_t	*symtab;
  elfshsect_t	*list;
  Elf32_Shdr	hdr;
  Elf32_Sym	new;
  Elf32_Sym	*actual;
  u_int		index;
  char		*str;

  if (file == NULL)
    ELFSH_SETERROR("[libelfsh:fixup_symtab] Invalid NULL parameter\n", NULL);

  /* Create the table if it does not exist */
  if (file->secthash[ELFSH_SECTION_SYMTAB] == NULL)
    {
      XALLOC(symtab, sizeof (elfshsect_t), NULL);
      hdr = elfsh_create_shdr(0, SHT_SYMTAB, 0, 0, 0, 0, 0, 0, 0, 0);
      symtab->name = strdup(ELFSH_SECTION_NAME_SYMTAB);
      index = elfsh_insert_unmapped_section(file, symtab, hdr, NULL);
      if (index < 0)
	return (NULL);
      file->secthash[ELFSH_SECTION_SYMTAB] = symtab;
    }
  else
    symtab = file->secthash[ELFSH_SECTION_SYMTAB];

  /* Fill strindex */
  if (strindex != NULL)
    elfsh_get_section_by_name(file, 
			      ELFSH_SECTION_NAME_STRTAB, 
			      strindex, NULL, NULL);

  /* Walk to the sht and fixup the symbol table */
  for (list = file->sectlist; list != NULL; list = list->next)
    {

      /* Create STT_SECTION symbol */
      actual = elfsh_get_sym_from_shtentry(file, list->shdr);
      if (actual == NULL)
	{
	  new = elfsh_create_symbol(list->shdr->sh_addr, 
				    list->shdr->sh_size, 
				    STT_SECTION, 0, 0, list->index);
	  elfsh_insert_symbol(symtab, &new, list->name);
	}

      /* Fixup STT_SECTION symbol */
      else
	{
	  actual->st_size = list->shdr->sh_size;
	  str = elfsh_get_symbol_name(file, actual);
	  if (str == NULL || !*str)
	    actual->st_name = elfsh_insert_in_strtab(file, list->name);
	}
    }

  /* Fixup all 0 lenght non-section-typed symbols */
  actual = symtab->data;
  for (index = 0; 
       index < symtab->shdr->sh_size / sizeof(Elf32_Sym); 
       index++, actual++)
    if (actual->st_value && !actual->st_size && 
	index + 1 < symtab->shdr->sh_size / sizeof(Elf32_Sym))
      actual->st_size = actual[1].st_value - actual->st_value;

  elfsh_sync_sorted_symtab(symtab);
  return (symtab);
}



/* Rebuild symbol string table if unexistant */
elfshsect_t	*elfsh_rebuild_strtab(elfshobj_t *file)
{
  elfshsect_t	*strtab;
  Elf32_Shdr	hdr;

  /* Sanity checks */
  if (file == NULL)
    ELFSH_SETERROR("[libelfsh:rebuild_strtab] Invalid NULL parameter\n", NULL);

  /* Create the table if it does not exist */
  if (file->secthash[ELFSH_SECTION_STRTAB] == NULL)
    {
      XALLOC(strtab, sizeof (elfshsect_t), NULL);
      hdr = elfsh_create_shdr(0, SHT_STRTAB, 0, 0, 0, 0, 0, 0, 0, 0);
      strtab->name = strdup(ELFSH_SECTION_NAME_STRTAB);
      elfsh_insert_unmapped_section(file, strtab, hdr, NULL);
      file->secthash[ELFSH_SECTION_STRTAB] = strtab;
    }
  return (file->secthash[ELFSH_SECTION_STRTAB]);
}









