/*
** plt.c for libelfsh
** 
** Started on  Tue Jul 24 06:40:07 2001 mayhem
** Last update Wed Apr 16 19:05:11 2003 mayhem
*/
#include "libelfsh.h"



/* Load the Procedure Linkage Table */
elfshsect_t	*elfsh_get_plt(elfshobj_t *file, int *num)
{
  elfshsect_t	*actual;
  void		*ret;

  if (file->sht == NULL && elfsh_get_sht(file, NULL) == NULL)
    return (NULL);

  for (actual = file->sectlist; actual != NULL; actual = actual->next)
    if (elfsh_is_plt(file, actual))
      {
	ret = elfsh_get_anonymous_section(file, actual);
	file->secthash[ELFSH_SECTION_PLT] = actual;
	return (ret ? actual : NULL);
      }
  
  return (NULL);
}


/* Return 1 if the section is the plt or 0 if its not */
int		elfsh_is_plt(elfshobj_t *file, elfshsect_t *sct)
{
  char		*name;

  name = elfsh_get_section_name(file, sct);
  if (name != NULL && !strcmp(name, ELFSH_SECTION_NAME_PLT))
    return (1);
  return (0);
}

/* Return 1 if the symbol points in the PLT, 0 if not, -1 if the PLT cant be found */
int			elfsh_is_pltentry(elfshobj_t *file, Elf32_Sym *sym)
{
  if (sym == NULL || elfsh_get_plt(file, NULL) == NULL)
    ELFSH_SETERROR("[libelfsh:is_pltentry] Invalid NULL parameter\n", -1);
  if (sym->st_size != file->secthash[ELFSH_SECTION_PLT]->shdr->sh_size)
    if (INTERVAL(file->secthash[ELFSH_SECTION_PLT]->shdr->sh_addr, 
		 sym->st_value, 
		 file->secthash[ELFSH_SECTION_PLT]->shdr->sh_addr + 
		 file->secthash[ELFSH_SECTION_PLT]->shdr->sh_size))
      return (1);
  return (0);
}


/* Return the size of 1 PLT entry */
int		elfsh_get_pltentsz(elfshobj_t *file)
{
  if (file == NULL)
    ELFSH_SETERROR("[libelfsh:get_pltentsz] Invalid NULL parameter\n", -1);
  switch (elfsh_get_arch(file->hdr))
    {
    case EM_386:
      return (16);
    case EM_SPARC:
    case EM_SPARC32PLUS:
      return (12);
    default:
      ELFSH_SETERROR("[libelfsh:get_pltentsz] Unsupported architecture\n", -1);
    }
}




























