/*
** relinject.c for libelfsh
** 
** Started on  Fri Mar 28 14:55:37 2003 mayhem
** Last update Mon May 19 14:09:12 2003 mayhem
*/
#include "libelfsh.h"

#define IS_SPARC(obj)   (((elfsh_get_arch((obj)->hdr) == EM_SPARC) || \
                          (elfsh_get_arch((obj)->hdr) == EM_SPARC32PLUS) || \
                          (elfsh_get_arch((obj)->hdr) == EM_SPARCV9)) ? 1 : 0)
#define IS_REL(scn)     (IS_SPARC((scn)->parent) ? 0 : 1)



/* Perform relocation on entry for INTEL architecture */
static int      elfsh_relocate_i386(elfshsect_t		*new, 
				     Elf32_Rel		*cur, 
				     u_long		*dword, 
				     u_long		addr)
{
  switch (elfsh_get_reltype(cur))
    {
      
      /* Absolute reference */
    case R_386_32:
      *dword = addr;
      break;
      
      /* Relative reference */
    case R_386_PC32:
      *dword = addr - (new->shdr->sh_addr + cur->r_offset + sizeof(u_long));
      break;
	  
    default:
      ELFSH_SETERROR("[libelfsh:relocate_etrel_section] "
		     "Unsupported relocation type\n", -1);
    }
  return (0);
}



/* Perform relocation on entry for SPARC architecture */
static int       elfsh_relocate_sparc(elfshsect_t       * new,
                                      Elf32_Rela        * cur,
                                      u_long            * dword,
                                      u_long              addr)
{
  int retval;
  u_long result;
  Elf32_Sym *sym;

#define ADD     (cur->r_addend)
#define BAS     (elfsh_get_object_baseaddr(new->parent))
#define GOT     0 /* GOT isn't available ... */
#define PLT     0 /* PLT isn't available ... */
#define PLA     (new->shdr->sh_addr + cur->r_offset)
#define SYM     (sym->st_value)

#define VERIFY8(x)      (((x) & (0xFFFFFFFF <<  8)) ? 1 : 0)
#define VERIFY13(x)     (((x) & (0xFFFFFFFF << 13)) ? 1 : 0)
#define VERIFY16(x)     (((x) & (0xFFFFFFFF << 16)) ? 1 : 0)
#define VERIFY22(x)     (((x) & (0xFFFFFFFF << 22)) ? 1 : 0)
#define VERIFY30(x)     (((x) & (0xFFFFFFFF << 30)) ? 1 : 0)
#define VERIFY32(x)     (((x) & (0xFFFFFFFF <<  0)) ? 1 : 0)

#define TRUNCATE8(x)    ((x) & ~(0xFFFFFFFF <<  8))
#define TRUNCATE13(x)   ((x) & ~(0xFFFFFFFF << 13))
#define TRUNCATE16(x)   ((x) & ~(0xFFFFFFFF << 16))
#define TRUNCATE22(x)   ((x) & ~(0xFFFFFFFF << 22))
#define TRUNCATE30(x)   ((x) & ~(0xFFFFFFFF << 30))
#define TRUNCATE32(x)   (x) /* XXX just assume its 32bit */
  
  sym = elfsh_get_symbol_from_reloc(new->parent, (Elf32_Rel *)cur);
  if (NULL == sym) 
    {
      // might not be the end of the world.. we don't always need sym.
    }

  retval = result = 0;
  switch (elfsh_get_reltype((Elf32_Rel *)cur))
    {
    case R_SPARC_NONE:
      break;
    case R_SPARC_8:
      result = SYM + ADD;
      if (VERIFY8(result))
	; // XXX error condition.
      break;
    case R_SPARC_16:
      result = SYM + ADD;
      if (VERIFY16(result))
	; // XXX error condition
      break;
    case R_SPARC_32:
      result = SYM + ADD;
      if (VERIFY32(result))
	; // XXX error condition
      break;
    case R_SPARC_DISP8:
      result = SYM + ADD - PLA;
      if (VERIFY8(result))
	; // XXX error condition
      break;
    case R_SPARC_DISP16:
      result = SYM + ADD - PLA;
      if (VERIFY16(result))
	; // XXX error condition
      break;
    case R_SPARC_DISP32:
      result = SYM + ADD - PLA;
      if (VERIFY32(result))
	; // XXX error condition
      break;
    case R_SPARC_WDISP30:
      result = (SYM + ADD - PLA) >> 2;
      if (VERIFY30(result))
	; // XXX error condition
      break;
    case R_SPARC_WDISP22:
      result = (SYM + ADD - PLA) >> 10;
      if (VERIFY22(result))
	; // XXX error condition
      break;
    case R_SPARC_HI22:
      result = TRUNCATE22(((SYM + ADD) >> 10));
      break;
    case R_SPARC_22:
      result = SYM + ADD;
      if (VERIFY22(result))
	; // XXX error condition
      break;
    case R_SPARC_13:
      result = SYM + ADD;
      if (VERIFY13(result))
	; // XXX error condition
      break;
    case R_SPARC_LO10:
      result = TRUNCATE13(((SYM + ADD) & 0x3FF));
      break;
    case R_SPARC_GOT10:
    case R_SPARC_GOT13:
    case R_SPARC_GOT22:
      // unsuppoted
      retval = -1;
      break;
    case R_SPARC_PC10:
      result = TRUNCATE13(((SYM + ADD - PLA) & 0x3FF));
      break;
    case R_SPARC_PC22:
      result = (SYM + ADD - PLA) >> 10;
      if (VERIFY22(result))
	; // XXX error condition
      break;
    case R_SPARC_WPLT30:
      // unsupported
      retval = -1;
      break;
    case R_SPARC_COPY:
      break;
    case R_SPARC_GLOB_DAT:
      result = SYM + ADD;
      if (VERIFY32(result))
	; // XXX error condition
      break;
    case R_SPARC_JMP_SLOT:
      retval = -1;
      break;
    case R_SPARC_RELATIVE:
      result = BAS + ADD;
      if (VERIFY32(result))
	; // XXX error condition
      break;
    case R_SPARC_UA32:
      result = SYM + ADD;
      if (VERIFY32(result))
	; // XXX error condition
      break;
    default:
      ELFSH_SETERROR("[libelfsh:relocate_sparc] Unsupported architecture\n",
		     -1);
      break;
    }
#undef ADD
#undef BAS
#undef GOT
#undef PLT
#undef PLA
#undef SYM
  *dword += result;
  return retval;
}




/* Perform relocation on entry */
static int      elfsh_relocate_entry(elfshsect_t        *new, 
                                     void               *reloc, 
                                     u_long             *dword, 
                                     u_long             addr)
{
  int retval;

  switch (elfsh_get_arch(new->parent->hdr))
    {
    case EM_386:
      /* case EM_486: XXX */
      retval = elfsh_relocate_i386(new, (Elf32_Rel *) reloc, dword, addr);
      break;
    case EM_SPARC:
    case EM_SPARC32PLUS:
      retval = elfsh_relocate_sparc(new, (Elf32_Rela *) reloc, dword, addr);
      break;
    case EM_SPARCV9: /* can't support just yet.. */
    default:
      ELFSH_SETERROR("[libelfsh:relocate_etrel_section] "
                     "Unsupported architecture\n", -1);
      break;
    }
  return retval;
}



/* Relocate the just injected section */
static int	elfsh_relocate_etrel_section(elfshsect_t *new, 
					     elfshsect_t *reltab)
{
  Elf32_Rel	*cur;
  Elf32_Sym	*sym;
  u_int		size;
  u_int		index;
  long		*dword;
  long		addr;
  char		*name;
  char		tmpname[BUFSIZ];
  elfshsect_t	*sect;

  /* Loop on the relocation table entries */
  size = reltab->shdr->sh_size / ELFSH_REL_ENTRY_SIZE;
  cur = reltab->data;
  for (index = 0; index < size; index++)
    {

      /* Get symbol value in ET_REL */
      sym  = elfsh_get_symbol_from_reloc(reltab->parent, cur + index);
      name = elfsh_get_symname_from_reloc(reltab->parent, cur + index);
      if (sym == NULL || name == NULL)
	return (-1);

      /* 
      ** If symbol type is NOTYPE, read value of existing symbol in 
      ** the ET_EXEC obj .
      **
      ** If symbol link is COMMON, we use the symbol inserted during 
      ** the ET_REL .bss sizescan in bss.c:elfsh_find_bsslen()
      */
      if (elfsh_get_symbol_type(sym) == STT_NOTYPE ||
	  elfsh_get_symbol_link(sym) == SHN_COMMON)
	{
	  sym = elfsh_get_metasym_by_name(new->parent, name);
	  if (sym == NULL)
	    ELFSH_SETERROR("[libelfsh:relocate_etrel_section] "
			   "Cant find requested symbol in ET_EXEC\n", -1);
	  addr = sym->st_value;

#if __DEBUG_RELADD__
	  printf("[DEBUG_RELADD] Relocating using existing sym. %-20s [%08X]\n",
		 name, (u_int) addr);
#endif

	}


      /* else, compute addr giving the injected section's vaddr in ET_EXEC */
      else
	{
	  
	  /* Find target section in ET_REL */
	  sect = elfsh_get_section_by_index(reltab->parent, sym->st_shndx,
					    NULL, NULL);
	  if (sect == NULL)
	    ELFSH_SETERROR("[libelfsh:relocate_etrel_section] "
			   "Cant find extracted section in ET_REL\n", -1);
	  
	  /* Find corresponding inserted section in ET_EXEC */
	  snprintf(tmpname, sizeof(tmpname), "%s%s", reltab->parent->name, sect->name);
	  sect = elfsh_get_section_by_name(new->parent, tmpname, NULL, NULL, NULL);
	  if (sect == NULL)
	    ELFSH_SETERROR("[libelfsh:relocate_etrel_section] "
			   "Cant find inserted section in ET_EXEC\n", -1);
	  addr = sect->shdr->sh_addr + sym->st_value;

#if __DEBUG_RELADD__
	  printf("[DEBUG_RELADD] Relocating by translated symbol %-20s [%08X]\n",
		 name, newsym.st_value);
#endif

	}


      /* Perform relocation */
      dword = (long *) ((char *) new->data + cur[index].r_offset);
      if (elfsh_relocate_entry(new, cur + index, dword, addr) < 0)
	return (-1);

    }
  return (0);
}



/* Inject a section from ET_REL object into ET_EXEC */
static int	elfsh_inject_etrel_section(elfshobj_t *file, elfshsect_t *sect)
{
  Elf32_Shdr	hdr;
  elfshsect_t	*new;
  elfshzone_t	*zone;
  char		*newname;
  char		writable;
  int		mode;
  char		*data;

  /* If the section is BSS, fuzion it with the existing BSS */
  if (sect->shdr->sh_type == SHT_NOBITS)
    {
      new = elfsh_fixup_bss(file);
      if (new == NULL)
	return (-1);
      zone = elfsh_create_bsszone(sect->parent->name, 
				  new->shdr->sh_size, 
				  sect->shdr->sh_size);
      elfsh_add_bsszone(new, zone);
      return (0);
    }

  /* else create a new section */
  hdr = elfsh_create_shdr(0, sect->shdr->sh_type, sect->shdr->sh_flags, 
			  0, 0, sect->shdr->sh_size, 0, 0, 0, 0);
  XALLOC(newname, strlen(sect->parent->name) + strlen(sect->name) + 2, -1);
  sprintf(newname, "%s%s", sect->parent->name, sect->name);
  new = elfsh_create_section(newname);

  /* Copy the data */
  XALLOC(data, sect->shdr->sh_size, -1);
  memcpy(data, sect->data, sect->shdr->sh_size);

  /* Inject new section by top or after bss depending on its type */
  writable = elfsh_get_section_writableflag(sect->shdr);
  mode     = (writable ? ELFSH_DATA_INJECTION : ELFSH_CODE_INJECTION);
  if (elfsh_insert_mapped_section(file, new, hdr, data, mode) < 0)
    goto bad;
  new = elfsh_get_section_by_name(file, newname, NULL, NULL, NULL);
  if (new == NULL)
    goto bad;

  return (0);
 bad:
  free(newname);
  return (-1);
}



/* Inject a ET_REL object into a ET_EXEC object */
int		elfsh_inject_etrel(elfshobj_t *file, elfshobj_t *rel)
{
  elfshsect_t	*sect;
  elfshsect_t	*reltab;
  char		sctname[BUFSIZ];
  Elf32_Sym	newsym;
  u_int		symnbr;
  u_int		index;
  static Elf32_Sym	*sym;
  u_char		type;

  /* Sanity checks */
  if (file == NULL || file->hdr == NULL || rel == NULL || rel->hdr == NULL)
    ELFSH_SETERROR("[libelfsh:inject_etrel] Invalid NULL parameter\n", -1);
  if (file->hdr->e_type != ET_EXEC || rel->hdr->e_type != ET_REL)
    ELFSH_SETERROR("[libelfsh:inject_etrel] Bad object types\n", -1);
  
  /* First pass : find and inject all allocatable sections */
  for (index = 0; index < rel->hdr->e_shnum; index++)
    {

      /* Get the current section */
      sect = elfsh_get_section_by_index(rel, index, NULL, NULL);
      if (sect == NULL)
	ELFSH_SETERROR("[libelfsh:inject_etrel] Cant read section in ET_REL\n", -1);

      /* Compute BSS lenght */
      if (sect->shdr->sh_type == SHT_NOBITS && elfsh_find_bsslen(file, sect) < 0)
	return (-1);

      /* Check if the current section need to be mapped */
      if (elfsh_get_section_allocflag(sect->shdr) && sect->shdr->sh_size &&
	  (sect->shdr->sh_type == SHT_NOBITS || 
	   sect->shdr->sh_type == SHT_PROGBITS) && 
	  elfsh_inject_etrel_section(file, sect) < 0)
      return (-1);
    }


#if __DEBUG_RELADD__
  printf("[DEBUG_RELADD] Entering intermediate symbol injection loop\n");
#endif


  /* Intermediate pass : Inject ET_REL symbol table into host file */
  sym = elfsh_get_symtab(rel, &symnbr);
  for (index = 0; index < symnbr; index++)
    {
      
      type = elfsh_get_symbol_type(sym + index);

      /* Avoid non-injectable symbols */
      if (type != STT_FUNC && type != STT_OBJECT)
	continue;
      if (sym[index].st_shndx >= rel->hdr->e_shnum)
	continue;

      /* Find target section in ET_REL */
      sect = elfsh_get_section_by_index(rel, sym[index].st_shndx, NULL, NULL);
      if (sect == NULL)
	ELFSH_SETERROR("[libelfsh:inject_etrel] "
		       "Cant find extracted section in ET_REL\n", -1);

      /* Avoid more symbols using source section .::. XXX: not sure it is necessary */
      if (sect->shdr->sh_type != SHT_PROGBITS || !sect->shdr->sh_size ||
	  !elfsh_get_section_allocflag(sect->shdr))
	continue;

      /* Find corresponding inserted section in ET_EXEC */
      snprintf(sctname, sizeof(sctname), "%s%s", rel->name, sect->name);
      sect = elfsh_get_section_by_name(file, sctname, NULL, NULL, NULL);
      if (sect == NULL)
	ELFSH_SETERROR("[libelfsh:inject_etrel] "
		       "Cant find inserted section in ET_EXEC\n", -1);
      
#if __DEBUG_RELADD__
      printf("[DEBUG_RELADD] Regulary injected symbol %-20s [%08X] \n", 
	     elfsh_get_symbol_name(rel, sym + index),
	     sect->shdr->sh_addr + sym[index].st_value);
#endif

      /* Add symbol in host file */
      newsym = elfsh_create_symbol(sect->shdr->sh_addr + sym[index].st_value, 
				   sym[index].st_size, 
				   elfsh_get_symbol_type(sym + index),
				   elfsh_get_symbol_bind(sym + index),
				   0, sect->index);
      if (elfsh_insert_symbol(file->secthash[ELFSH_SECTION_SYMTAB], &newsym,
			      elfsh_get_symbol_name(rel, sym + index)) < 0)
	return (-1);
    }



#if __DEBUG_RELADD__
  printf("[DEBUG_RELADD] Entering final relocation loop\n");
#endif



  /* Second pass : relocate each inserted section */
  for (index = 0; index < rel->hdr->e_shnum; index++)
    {
      sect = elfsh_get_section_by_index(rel, index, NULL, NULL);
      if (sect == NULL)
	ELFSH_SETERROR("[libelfsh:inject_etrel] Cant get section in ET_REL\n", -1);

      /* Check if the section is mapped */
      if (elfsh_get_section_allocflag(sect->shdr) && sect->shdr->sh_size &&
	  sect->shdr->sh_type == SHT_PROGBITS)
	{
	  
	  /* Find the associate .rel section */
	  snprintf(sctname, sizeof(sctname), "%s%s", 
		   (IS_REL(sect) ? ".rel" : ".rela"), sect->name);
	  reltab = elfsh_get_section_by_name(rel, sctname, NULL, NULL, NULL);
	  if (reltab == NULL)
	    continue;

	  /* Find the injected instance of this allocatable section in the ET_EXEC */
	  snprintf(sctname, sizeof(sctname), "%s%s", sect->parent->name, sect->name);
	  sect = elfsh_get_section_by_name(file, sctname, NULL, NULL, NULL);
	  if (sect == NULL)
	    ELFSH_SETERROR("[libelfsh:inject_etrel] Cant get section in ET_EXEC\n", -1);
	  if (elfsh_relocate_etrel_section(sect, reltab) < 0)
	    return (-1);
	}
    }
  return (0);
}
















